import re

from django.urls.resolvers import RoutePattern, _route_to_regex

from garpix_page.cache import cache_service
from garpix_page.utils.get_current_language_code_url_prefix import get_current_language_code_url_prefix
from django.utils.translation import activate

from garpix_page.utils.get_garpix_page_models import get_garpix_page_models


class PageViewMixin:

    @staticmethod
    def get_absolute_url_from_request(slug_list, slug):
        current_language_code_url_prefix = get_current_language_code_url_prefix()
        if slug != '':
            return "{}/{}".format(current_language_code_url_prefix, '/'.join(slug_list))
        return "{}".format(current_language_code_url_prefix) if len(current_language_code_url_prefix) > 1 else '/'

    @classmethod
    def get_instance_by_slug(cls, slugs, languages_list):
        slug_list = slugs.split('/')

        lang = None

        if slug_list[0] in languages_list:
            lang = slug_list.pop(0)
            activate(lang)

        if len(slug_list) == 0:
            slug_list.append('')

        if len(slug_list) > 1 and slug_list[-1] == '':
            slug_list.pop(-1)

        slug = slug_list[-1]

        url = cls.get_absolute_url_from_request(slug_list, slug)

        instance_cache = cache_service.get_instance_by_url(url)

        if instance_cache is not None:
            return instance_cache

        page_models = get_garpix_page_models()
        active_models = []
        for el in page_models:
            for key, value in el.url_patterns().items():
                pattern = RoutePattern(r'(<url>){}'.format(value['pattern']))
                pattern.regex = re.compile('^/(?P<url>.*){}$'.format(_route_to_regex(value['pattern'], pattern._is_endpoint)[0][1:]))
                match = pattern.match(url)
                if match is not None:
                    last, _, params = match

                    el_url = f"/{params['url']}"
                    el_slug = '' if lang and params['url'] == lang else el_url.split('/')[-1]

                    active_models.append({
                        'model': el,
                        'params': params,
                        'pattern': key,
                        'url': el_url,
                        'slug': el_slug
                    })

        for model in active_models:
            instances = model['model'].active_on_site.filter(slug=model['slug']).all()

            for instance in instances:

                if instance.absolute_url == model['url']:
                    instance.subpage_params = model['params']
                    instance.subpage_key = model['pattern']
                    cache_service.set_instance_by_url(url, instance)
                    return instance

        return None



from django.db import models
from garpix_utils.models import GarpixSiteConfiguration


class GarpixPageSiteConfiguration(GarpixSiteConfiguration):
    class CHANGEFRAQ:
        ALWAYS = 'always'
        HOURLY = 'hourly'
        DAILY = 'daily'
        WEEKLY = 'weekly'
        MONTHLY = 'monthly'
        YEARLY = 'yearly'
        NEVER = 'never'

        CHOICES = (
            (ALWAYS, 'always'),
            (HOURLY, 'hourly'),
            (DAILY, 'daily'),
            (WEEKLY, 'weekly'),
            (MONTHLY, 'monthly'),
            (YEARLY, 'yearly'),
            (NEVER, 'never')
        )

    robots_txt = models.TextField(verbose_name='Содержимое файла robots.txt',
                                  default="User-agent: *\nDisallow: /admin/\nDisallow: /api/")
    sitemap_frequency = models.CharField(max_length=7, choices=CHANGEFRAQ.CHOICES, default=CHANGEFRAQ.ALWAYS,
                                         verbose_name='Sitemap changefreq')

    class Meta:
        verbose_name = 'Настройки'
        verbose_name_plural = 'Настройки'



from django.contrib import admin
from django.utils.html import format_html

from .forms import PageForm
from ..models.base_page import BasePage
from modeltranslation.admin import TabbedTranslationAdmin
from polymorphic_tree.admin import PolymorphicMPTTParentModelAdmin, PolymorphicMPTTChildModelAdmin

from django.urls import reverse
from django.http import HttpResponseRedirect
from django.urls import path

from ..models.components.base_component import PageComponent
from ..utils.all_sites import get_all_sites
from ..utils.get_garpix_page_models import get_garpix_page_models
from django.conf import settings
from django.utils.translation import gettext as _
from polymorphic.admin import PolymorphicChildModelFilter
from tabbed_admin import TabbedModelAdmin
from mptt.admin import DraggableMPTTAdmin
from garpix_admin_lock.mixins import PageLockAdminMixin

from ..utils.get_languages import get_languages
from ..views.clear_cache import clear_cache


class ComponentsTabularInline(admin.TabularInline):
    model = PageComponent
    fields = ('component', 'view_order')
    raw_id_fields = ('component',)
    extra = 0


class PageAdmin(TabbedModelAdmin, TabbedTranslationAdmin, PolymorphicMPTTChildModelAdmin):
    base_model = BasePage
    list_per_page = settings.GARPIX_PAGE_ADMIN_LIST_PER_PAGE if hasattr(settings,
                                                                        'GARPIX_PAGE_ADMIN_LIST_PER_PAGE') else 25
    empty_value_display = '- нет -'
    save_on_top = True
    view_on_site = True

    base_form = PageForm

    change_form_template = 'garpix_page/admin/page_change_form.html'

    date_hierarchy = 'created_at'
    prepopulated_fields = {'slug': ('title',)}

    search_fields = ('title',)
    list_filter = ('is_active', 'created_at', 'updated_at')
    actions = ('clone_object', 'rebuild')

    list_display = ('title', 'created_at', 'is_active', 'get_absolute_url',)
    list_editable = ('is_active',)
    raw_id_fields = ('parent',)

    readonly_fields = ('get_absolute_url', 'created_at', 'updated_at')

    def get_form(self, request, *args, **kwargs):
        form = super().get_form(request, *args, **kwargs)
        obj = args[0]

        if obj and isinstance(obj, BasePage):
            sites = obj.sites.all() if obj.sites else get_all_sites()
            lang_seo_fields = ['seo_title', 'seo_keywords', 'seo_description', 'seo_author']
            non_lang_seo_fields = ['seo_og_type', 'seo_image']
            for field in non_lang_seo_fields:
                form.base_fields[field].help_text = "Итоговое значение:"
                for site in sites:
                    form.base_fields[field].help_text += f"<br> для сайта {site.name} - {obj.get_seo_value(field_name=field, site=site)}"

            for field in lang_seo_fields:
                for lang in get_languages():
                    lang = lang.replace('-', '_')
                    form.base_fields[f'{field}_{lang}'].help_text = "Итоговое значение:"
                    for site in sites:
                        form.base_fields[f'{field}_{lang}'].help_text += f"<br> для сайта {site.name} - {obj.get_seo_value(field_name=f'{field}_{lang}', site=site)}"

        # form.current_user = request.user
        return form

    def delete_queryset(self, request, queryset):
        self.model.objects.delete(id__in=queryset.values_list('id', flat=True))
        self.model.objects.rebuild()

    def get_actions(self, request):
        actions = super().get_actions(request)
        if actions is not None and "delete_selected" in actions:
            actions["delete_selected"] = (
                self.delete_queryset,
                "delete_selected",
                _("Delete selected %(verbose_name_plural)s"),
            )
        return actions

    def has_module_permission(self, request):
        return False

    def get_fieldsets(self, request, obj=None):
        """
        Если в модели страницы определены табы, будут отображаться они.
        Если не определены - все доступные филдсеты/поля будут помещены в таб "Основное", сео теги - в таб "Сео"
        """
        if self.tabs is None:
            fields = self.get_fields(request, obj)
            tab_seo_fields = []
            tab_main_fields = []
            for field in fields:
                if field[:4] == 'seo_':
                    tab_seo_fields.append(field)
                else:
                    tab_main_fields.append(field)
            tab_seo = (
                (None, {
                    'fields': tab_seo_fields
                }),
            )

            if self.fieldsets:
                tab_main = self.fieldsets
            else:
                tab_main = (
                    (None, {
                        'fields': tab_main_fields
                    }),
                )

            if self.inlines:
                tab_main += tuple(self.inlines)

            self.tabs = [
                ('Основное', tab_main),
                ('SEO', tab_seo),
                ('Компоненты', (ComponentsTabularInline,))
            ]
        tabs_fieldsets = self.get_formatted_tabs(request, obj)['fieldsets']
        self.fieldsets = ()
        self.fieldsets = self.add_tabbed_item(tabs_fieldsets, self.fieldsets)
        return super().get_fieldsets(request, obj)


@admin.register(BasePage)
class RealPageAdmin(DraggableMPTTAdmin, TabbedTranslationAdmin, PolymorphicMPTTParentModelAdmin):
    """
    Стандартные настройки для базовых страниц.
    """
    base_model = BasePage
    child_models = get_garpix_page_models()

    list_per_page = settings.GARPIX_PAGE_ADMIN_LIST_PER_PAGE if hasattr(settings,
                                                                        'GARPIX_PAGE_ADMIN_LIST_PER_PAGE') else 25

    empty_value_display = '- нет -'
    save_on_top = True
    view_on_site = True

    date_hierarchy = 'created_at'
    prepopulated_fields = {'slug': ('title',)}

    search_fields = ('title',)
    list_filter = (PolymorphicChildModelFilter, 'is_active', 'created_at', 'updated_at', 'sites')
    actions = ('clone_object', 'rebuild')

    list_display = ('tree_actions', 'indented_title', 'created_at', 'is_active',
                    'get_absolute_url_html_admin', 'model_name')
    list_editable = ('is_active',)

    readonly_fields = ('created_at', 'updated_at', 'model_name')

    def indented_title(self, item):
        return super().indented_title(item)

    indented_title.short_description = "Название"

    def get_absolute_url_html_admin(self, obj):
        return format_html('<a href="{0}" target="_blank">{0}</a>', obj.absolute_url)

    get_absolute_url_html_admin.short_description = 'URL'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if 'initial' in kwargs:
            self.fields['parent'].queryset = self.model.get_available_page_parents()
            self.fields['parent'].default = self.model.get_default_parent()

    def clone_object(self, request, queryset):
        """Копирование(клонирование) выбранных объектов - action"""
        for obj in queryset:
            obj = obj.get_real_instance()

            len_old_title = obj.__class__.objects.filter(title__icontains=obj.title).count()
            title = f"{obj.title} ({len_old_title})"
            slug = f"{obj.slug}-{len_old_title}"
            obj.title = title
            obj.slug = slug
            new_obj = obj.clone_object(title=title, slug=slug)
            new_obj.save()

    clone_object.short_description = 'Клонировать объект'

    def _rebuild(self):
        try:
            self.model.objects.rebuild()
        except:  # noqa
            print('[ERROR]: Ошибка при перезагрузки древовидной структуры')

    def rebuild(self, request, queryset):
        """Пересорбать МПТТ модель. Иногда требуется для перезагрузки дерева."""
        self._rebuild()

    rebuild.short_description = 'Пересобрать пункты раздела'

    def save_model(self, request, obj, form, change):
        self._rebuild()
        super().save_model(request, obj, form, change)

    def get_urls(self):
        urls = super().get_urls()

        info = self.model._meta.app_label, self.model._meta.model_name

        my_urls = [
            path('<path:pk>/full_clone/', self.full_clone, name='%s_%s_full_clone' % info),
            path('admin_clear_cache', clear_cache, name='admin_clear_cache')
        ]

        return my_urls + urls

    def full_clone(self, request, pk):
        if request.method == 'POST':
            obj = self.get_object(request, pk)
            obj = obj.get_real_instance()
            title = request.POST.get('title', None)
            len_old_title = obj.__class__.objects.filter(title__icontains=obj.title).count()
            if not title:
                title = f"{obj.title} ({len_old_title})" if len_old_title > 0 else obj.title
            slug = f"{obj.slug}-{len_old_title}"
            new_obj = obj.clone_object(title=title, slug=slug)
            new_obj.save()
        link = reverse("admin:garpix_page_basepage_changelist")
        return HttpResponseRedirect(link)


#  Базовая админка для страниц с локом
class BasePageAdmin(PageLockAdminMixin, PageAdmin):
    lock_change_view = True
    change_form_template = 'garpix_page/admin/page_change_form.html'

    def has_change_permission(self, req, *args):
        can_add = super().has_change_permission(req, *args)

        if can_add and not self._is_locked(req):
            return True

        return False


class RealBasePageAdmin(PageLockAdminMixin, RealPageAdmin):
    lock_change_view = True


#  Базовая админка для страниц без лока
class RealBasePageWithoutLockAdmin(PageAdmin):
    pass


class BasePageWithoutLockAdmin(PageAdmin):
    pass



from .get_languages import get_languages


def get_exclude_fields(model):
    langs = get_languages()
    exclude_fields = []
    model_fields = [field.name for field in model._meta.get_fields()]

    for field in model_fields:
        for lang in langs:
            if f'{field}_{lang.replace("-", "_")}' not in model_fields:
                break
        else:
            for lang in langs:
                exclude_fields.append(f'{field}_{lang.replace("-", "_")}')
    return exclude_fields



from modeltranslation.translator import TranslationOptions, register
from garpix_page.models import BaseComponent


@register(BaseComponent)
class BaseComponentTranslationOptions(TranslationOptions):
    fields = ('title', 'text_title')



from django import template
from django.utils.safestring import mark_safe
from garpix_page import settings


register = template.Library()


@register.simple_tag
def get_settings_value(name, safe=False):
    if not safe:
        return getattr(settings, name)

    return mark_safe(str(getattr(settings, name)))



from django.contrib import messages
from django.http import HttpResponseRedirect
from garpix_page.cache import cache_service
from django.urls import reverse


def clear_cache(request):
    if not request.user.is_authenticated or not request.user.is_staff:
        return HttpResponseRedirect(reverse('admin:index'))
    cache_service.clear_all()
    link = request.META.get('HTTP_REFERER', reverse('admin:index'))
    messages.add_message(request, messages.SUCCESS, 'Cache has been cleared', extra_tags='cache_cleared')
    return HttpResponseRedirect(link)



from django.conf import settings
from django.utils.module_loading import import_string

from garpix_page.cache import cache_service
from garpix_page.utils.get_current_language_code_url_prefix import get_current_language_code_url_prefix

celery_app = import_string(settings.GARPIXCMS_CELERY_SETTINGS)


@celery_app.task()
def clear_child_cache(instance_pk):
    from garpix_page.models import BasePage

    instance = BasePage.objects.get(pk=instance_pk)

    pages = BasePage.objects.get_queryset_descendants(instance.get_children(), include_self=True)

    for page in pages:
        cache_service.reset_url_info_by_page(page, get_current_language_code_url_prefix())



from collections import OrderedDict

from rest_framework.pagination import PageNumberPagination


class GarpixPagePagination(PageNumberPagination):
    page_size = 20
    page_size_query_param = 'page_size'

    def get_paginated_data(self, queryset, serializer, request):

        page = self.paginate_queryset(queryset, request)

        if page is not None:
            serializer = serializer(page, context={'request': request}, many=True)
            return self.get_paginated_response(serializer.data)
        else:
            return serializer(queryset, many=True).data

    def get_paginated_response(self, data):
        return OrderedDict([
            ('count', self.page.paginator.count),
            ('next', self.get_next_link()),
            ('previous', self.get_previous_link()),
            ('results', data)
        ])



from django.contrib.sites.models import Site
from django.db import models
from garpix_utils.file import get_file_path
from django.utils.translation import gettext as _
from garpix_utils.managers import GCurrentSiteManager
from garpix_utils.models import ActiveMixin

from garpix_page.utils.all_sites import get_all_sites
from garpix_page.cache import cache_service


class SeoTemplate(ActiveMixin, models.Model):
    rule_field = models.CharField(verbose_name=_('Поле'), max_length=255)

    model_rule_value = models.CharField(verbose_name=_('Название'), null=True, blank=True, max_length=255)
    rule_value = models.CharField(verbose_name=_('Значение'), null=True, blank=True, max_length=255)

    seo_title = models.CharField(max_length=250, verbose_name=_('SEO заголовок страницы (title)'), blank=True,
                                 default='')
    seo_keywords = models.CharField(max_length=250, verbose_name=_('SEO ключевые слова (keywords)'), blank=True,
                                    default='')
    seo_description = models.TextField(verbose_name=_('SEO описание (description)'), blank=True, default='')
    seo_author = models.CharField(max_length=250, verbose_name=_('SEO автор (author)'), blank=True, default='')
    seo_og_type = models.CharField(max_length=250, verbose_name=_('SEO og:type'), blank=True, default="website")
    seo_image = models.FileField(upload_to=get_file_path, blank=True, null=True, verbose_name=_('SEO изображение'))

    priority_order = models.PositiveIntegerField(default=1, help_text=_('Чем меньше число, тем выше приоритет'),
                                                 verbose_name=_('Приоритетность применения'))
    sites = models.ManyToManyField(Site, default=get_all_sites, verbose_name='Сайты для применения')

    objects = models.Manager()
    on_site = GCurrentSiteManager()

    class Meta:
        verbose_name = 'Шаблон для seo'
        verbose_name_plural = 'Шаблоны для seo'
        ordering = ['priority_order', 'id']

    def __str__(self):
        return f'Шаблон для seo # {self.id}'

    def save(self, force_insert=False, force_update=False, using=None,
             update_fields=None):
        super().save()
        cache_service.clear_seo_data()



from .base_page import (  # noqa
    BasePageAdmin,
    RealBasePageAdmin,
    RealBasePageWithoutLockAdmin,
    BasePageWithoutLockAdmin,
)
from .components.grapes_js_html import GrapesJsHtmlComponentAdmin  # noqa
from .settings import *  # noqa



from django import template

register = template.Library()


@register.simple_tag(takes_context=True)
def gx_component(context, component):
    user = context['user']
    component_path = component['admin_edit_url']

    return f'data-gx-component={component_path}' if user.is_authenticated and user.is_staff else ''



from django.http import HttpResponse
from django.views.decorators.http import require_GET

from garpix_page.models import GarpixPageSiteConfiguration


@require_GET
def robots_txt(request):

    try:
        lines = GarpixPageSiteConfiguration.get_solo().robots_txt
    except Exception:
        lines = ''
    return HttpResponse(lines, content_type="text/plain")



from django.http import HttpResponse
from django.views.generic import TemplateView
from garpix_page.utils import get_render_html_value

__all__ = ('GetTemplate', )


class GetTemplate(TemplateView):
    def get(self, request, *args, **kwargs):
        template_name = request.GET['template_name']
        apply_django_tag = int(request.GET['apply_django_tag'])

        return HttpResponse(get_render_html_value(template_name, apply_django_tag=apply_django_tag)())



from django.conf import settings


# path to the html file of the form field. Enter your value for the override
GRAPESJS_TEMPLATE = getattr(settings, 'GRAPESJS_TEMPLATE', 'garpix_page/grapesjs/textarea.html')

# True if you want to save html and css
GRAPESJS_SAVE_CSS = int(getattr(settings, 'GRAPESJS_SAVE_CSS', False))

GRAPESJS_SAVE_ASSETS_TO_FILE = int(getattr(settings, 'GRAPESJS_SAVE_ASSETS_TO_FILE', False))

# use the value of the field from the db - True, or use the global save editor
GRAPESJS_DEFAULT_MODELS_DATA = int(getattr(settings, 'GRAPESJS_DEFAULT_MODELS_DATA', True))

# redefine the path to the html file, the markup from this file will be used by default
GRAPESJS_DEFAULT_HTML = getattr(settings, 'GRAPESJS_DEFAULT_HTML', 'garpix_page/grapesjs/default.html')

MIN = 'min'
BASE = 'base'

# Use this dictionary to override or add editor configurations
REDACTOR_CONFIG = {
    MIN: 'garpix_page/grapesjs/redactor_config/min.html',
    BASE: 'garpix_page/grapesjs/redactor_config/base.html',
    **getattr(settings, 'REDACTOR_CONFIG', {})
}

TEMPLATE_DIR = getattr(settings, 'TEMPLATE_DIR', 'templates')
STATIC_URL = getattr(settings, 'STATIC_URL', "/static/")

# you can override the name of tags
NAME_IGNORE_TAG = getattr(settings, 'NAME_IGNORE_TAG', 'ignore')
NAME_HIDDEN_TAG = getattr(settings, 'NAME_HIDDEN_TAG', 'hidden')
NAME_RENDER_TAG = getattr(settings, 'NAME_RENDER_TAG', 'render')
NAME_MAKEUP_TAG = getattr(settings, 'NAME_MAKEUP_TAG', 'makeup')


REPLACE_SAVE_IGNORE_TAGS = {
    '<%s>' % NAME_IGNORE_TAG: '{# <%s>' % NAME_IGNORE_TAG,
    '</%s>' % NAME_IGNORE_TAG: '</%s> #}' % NAME_IGNORE_TAG,
}
REPLACE_INIT_IGNORE_TAGS = {
    '{# <%s>' % NAME_IGNORE_TAG: '<%s>' % NAME_IGNORE_TAG,
    '</%s> #}' % NAME_IGNORE_TAG: '</%s>' % NAME_IGNORE_TAG,
}

REPLACE_SAVE_HIDDEN_TAGS = {
    '<div hidden=""><%s>' % NAME_HIDDEN_TAG: '<%s>' % NAME_HIDDEN_TAG,
    '</%s></div>' % NAME_HIDDEN_TAG: '</%s>' % NAME_HIDDEN_TAG,
}
REPLACE_INIT_HIDDEN_TAGS = {
    '<%s>' % NAME_HIDDEN_TAG: '<div hidden=""><%s>' % NAME_HIDDEN_TAG,
    '</%s>' % NAME_HIDDEN_TAG: '</%s></div>' % NAME_HIDDEN_TAG,
}

STRING_HANDLERS = [
    'garpix_page.utils.tags.makeup.ApplyMakeupTag',
    'garpix_page.utils.tags.render.ApplyRenderTag',
    *getattr(settings, 'STRING_HANDLERS', [])
]



from django.template.backends.dummy import TemplateStrings
from garpix_page.settings import TEMPLATE_DIR

__all__ = ('template_source', )


TemplateStrings.app_dirname = TEMPLATE_DIR
template_source = TemplateStrings({'DIRS': [], 'APP_DIRS': True, 'OPTIONS': {}, 'NAME': 'django'})
